/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __draw_DrawCached_h__
#define __draw_DrawCached_h__

#define DCH_MAX_VERTS	512

namespace fe
{

/**************************************************************************//**
	@brief Delayed rendering to another DrawI, presumably an immediate one

	@ingroup draw

	Opportunites exist to use the cache for sorting, optimization,
	and off-threading.
*//***************************************************************************/
class FE_DL_EXPORT DrawCached:
	public DrawCommon,
	public Initialize<DrawCached>
{
	private:

	class Buffer
	{
		public:
			SpatialVector	m_pVertex[DCH_MAX_VERTS];
			SpatialVector	m_pNormal[DCH_MAX_VERTS];
			Color			m_pColor[DCH_MAX_VERTS];
			Vector2f		m_pTexture[DCH_MAX_VERTS];
#if FE_DOUBLE_REAL
			U32				m_pFiller[DCH_MAX_VERTS];
#endif
	};

	//* byte-wise same size as m_pTexture (for blind cast)
	class CylinderOverlay
	{
		public:
			Real			m_pBaseScale[DCH_MAX_VERTS];
			U32				m_pSlices[DCH_MAX_VERTS];
	};

	class Command
	{
		public:

			enum	Instruction
					{
						e_null,
						e_points,
						e_lines,
						e_triangles,
						e_rectangles,
						e_sphere,
						e_cylinder,
						e_cylinders,
						e_raster,
						e_drawable,
						e_text,
						e_max
					};

							Command(void):
								m_projection(ViewMode::e_current),
								m_instruction(e_null),
								m_stripMode(DrawI::e_discrete),
								m_multicolor(FALSE),
								m_vertices(0),
								m_pVertex(NULL),
								m_pNormal(NULL),
								m_pColor(NULL),
								m_pTexture(NULL)
							{
								m_pVertex=m_localBuffer.m_pVertex;
								m_pNormal=m_localBuffer.m_pNormal;
								m_pColor=m_localBuffer.m_pColor;
								m_pTexture=m_localBuffer.m_pTexture;
							}

							Command(const Command& rCommand)
							{	operator=(rCommand); }

			Command&		operator=(const Command& rCommand);

			I32				layer(void) const
							{	return m_spDrawMode.isValid()?
										m_spDrawMode->layer(): -1; }

			sp<ViewMode>			m_spViewMode;
			sp<DrawMode>			m_spDrawMode;
			sp<ImageI>				m_spImageI;
			cp<DrawableI>			m_cpDrawableI;
			ViewMode::Projection	m_projection;
			Instruction				m_instruction;
			StripMode				m_stripMode;
			BWORD					m_multicolor;
			U32						m_vertices;

			SpatialVector*			m_pVertex;
			SpatialVector*			m_pNormal;
			Color*					m_pColor;
			Vector2f*				m_pTexture;

									//* TEMP should use variable size pool
			Buffer					m_localBuffer;
	};

	class CommandQueue: public Array<Command>
	{
		public:
						CommandQueue(void):
							m_commands(0)
						{
							reserve(256);
						}

			Command&	access(U32 index)
						{
							if(index>=size())
							{
								resize(index+1);
							}
							return Array<Command>::operator[](index);
						}

			U32			m_commands;
	};

	public:
					DrawCached();
virtual				~DrawCached();

		void		initialize(void);

virtual	void		flush(void);
virtual	void		flushLive(void);
virtual	void		clearInput(void);
virtual	BWORD		empty(void) const;

virtual	BWORD		isDirect(void) const
					{	return m_spDrawChain.isValid() &&
								m_spDrawChain->isDirect(); }
virtual	void		setFontBase(U32 index)
					{	prepareChain();
						m_spDrawChain->setFontBase(index); }

virtual void		drawPoints(const SpatialVector *vertex,
							const SpatialVector *normal,U32 vertices,
							BWORD multicolor,const Color *color);

virtual void		drawLines(const SpatialVector *vertex,
							const SpatialVector *normal,U32 vertices,
							StripMode strip,BWORD multicolor,
							const Color *color);

virtual	void		drawTriangles(const SpatialVector *vertex,
							const SpatialVector *normal,
							const Vector2f *texture,U32 vertices,
							StripMode strip,BWORD multicolor,
							const Color *color);

virtual void		drawRectangles(const SpatialVector *vertex,U32 vertices,
						BWORD multicolor,const Color *color);

virtual void		drawSphere(const SpatialTransform &transform,
							const SpatialVector *scale,const Color &color);

virtual	void		drawCylinder(const SpatialTransform &transform,
							const SpatialVector *scale,
							Real baseScale,const Color &color,U32 slices);
virtual void		drawCylinders(const SpatialTransform* transform,
							const SpatialVector* scale,const Real *baseScale,
							const U32* slices,U32 cylinders,
							BWORD multicolor,const Color* color);

virtual	void		draw(cp<DrawableI> cpDrawableI,const Color* color);

virtual	void		drawTransformed(const SpatialTransform &transform,
							cp<DrawableI> cpDrawableI,const Color* color);

virtual	void		drawRaster(sp<ImageI> spImageI,
							const SpatialVector& location);

virtual	void		drawAlignedText(const SpatialVector &location,
							const String text,const Color &color);

virtual	void		setDrawChain(sp<DrawI> a_spDrawI)
					{	m_spDrawChain=a_spDrawI; }

	protected:

		BWORD			prepareChain(void);
		void			drawOutput(U32 queue,BWORD a_clearQueue);
		void			clearViewModeMap(void);

		sp<DrawI>		m_spDrawChain;
		U32				m_inputQueue;
		U32				m_outputQueue;
		U32				m_queues;

	private:

		void			drawGeometry(Command::Instruction instruction,
								const SpatialVector *vertex,
								const SpatialVector *normal,
								const Vector2f *texture,
								const Real* auxReal,const U32* auxU32,
								U32 vertices,StripMode strip,BWORD multicolor,
								const Color *color);

		void			adoptView(DrawCached::Command& rCommand,
								sp<ViewI> spViewI);

		U32				m_grain[Command::e_max];
		U32				m_oversize[Command::e_max];

		CommandQueue*	m_pCommandQueue;

		std::map<sp<ViewMode>, sp<ViewMode> >	m_viewModeMap;
		sp<WindowI>		m_spWindow;
};

} // namespace

#endif /* __draw_DrawCached_h__ */
