/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SAR_DEBUG			FALSE
#define FE_SAR_CLONE_DEBUG		FALSE

#include "surface.pmh"

namespace fe
{

SurfaceAccessibleRecord::SurfaceAccessibleRecord(void)
{
}

SurfaceAccessibleRecord::~SurfaceAccessibleRecord(void)
{
}

void SurfaceAccessibleRecord::reset(void)
{
	if(m_spScope.isValid())
	{
		SurfaceModel surfaceModelRV;
		surfaceModelRV.bind(m_spScope);
		m_weakRecord=surfaceModelRV.createRecord();

		FEASSERT(m_weakRecord.isValid());

		m_spRecordGroup=new RecordGroup();
		m_spRecordGroup->add(m_weakRecord);
	}
	else
	{
		//* TODO check this
		FEASSERT(FALSE);

		m_weakRecord=Record();
		m_spRecordGroup=NULL;
	}
}

Protectable* SurfaceAccessibleRecord::clone(Protectable* pInstance)
{
#if	FE_SAR_CLONE_DEBUG
	feLog("SurfaceAccessibleRecord::clone\n");

	const U32 us1=systemMicroseconds();
#endif

	FEASSERT(m_spScope.isValid());

	SurfaceAccessibleRecord* pSurfaceAccessibleRecord= pInstance?
			dynamic_cast<SurfaceAccessibleRecord*>(pInstance):
			new SurfaceAccessibleRecord();

	pSurfaceAccessibleRecord->setLibrary(library());

	//* NOTE would be risky to share scope

	sp<Scope> spScope=registry()->create("Scope");
	pSurfaceAccessibleRecord->bind(spScope);

	sp<RecordGroup>& rspRecordGroup=pSurfaceAccessibleRecord->m_spRecordGroup;

#if	FE_SAR_CLONE_DEBUG
	const U32 us2=systemMicroseconds();
#endif

	//* why is direct cloning slower?
//	Cloner cloner(m_spScope);
//	rspRecordGroup=cloner.clone(m_spRecordGroup);

	sp<data::StreamI> spStream=
			sp<data::StreamI>(new data::BinaryStream(m_spScope));

	std::stringstream stringStream;
	spStream->output(stringStream,m_spRecordGroup);

#if	FE_SAR_CLONE_DEBUG
	const U32 us3=systemMicroseconds();
#endif

	rspRecordGroup=spStream->input(stringStream);

#if	FE_SAR_CLONE_DEBUG
	const U32 us4=systemMicroseconds();
#endif

	//* update surface with new Record
	RecordGroup::iterator it=rspRecordGroup->begin();
	sp<RecordArray> spRA= *it;
	Record record=spRA->getRecord(0);
	pSurfaceAccessibleRecord->bind(record);

	FEASSERT(pSurfaceAccessibleRecord->m_spScope.isValid());

#if	FE_SAR_CLONE_DEBUG
	const U32 us5=systemMicroseconds();

	feLog("SurfaceAccessibleRecord::clone us %d %d %d %d -> %d\n",
			us2-us1,us3-us2,us4-us3,us5-us4,us5-us1);
#endif

	return pSurfaceAccessibleRecord;
}

BWORD SurfaceAccessibleRecord::load(String a_filename,sp<Catalog> a_spSettings)
{
#if	FE_SAR_DEBUG
	feLog("SurfaceAccessibleRecord::load \"%s\"\n",a_filename.c_str());
#endif

	BWORD binary=FALSE;

	if(a_spSettings.isValid())
	{
		const String options=a_spSettings->catalog<String>("options");
		binary=options.match(".*\\bbinary\\b.*");
	}

	FEASSERT(m_spScope.isValid());
	m_spRecordGroup=RecordView::loadRecordGroup(m_spScope,a_filename,binary);
	if(m_spRecordGroup.isNull())
	{
		feLog("SurfaceAccessibleRecord::load NULL RecordGroup\n");
		return FALSE;
	}

	RecordGroup::iterator it=m_spRecordGroup->begin();
	sp<RecordArray> spRA= *it;

	FEASSERT(spRA.isValid());
	m_weakRecord=spRA->getRecord(0);

	return TRUE;
}

BWORD SurfaceAccessibleRecord::save(String a_filename,sp<Catalog> a_spSettings)
{
#if	FE_SAR_DEBUG
	feLog("SurfaceAccessibleRecord::save\n");
#endif

	FEASSERT(m_spScope.isValid());
	FEASSERT(m_spRecordGroup.isValid());

	if(m_spScope.isNull() || m_spRecordGroup.isNull())
	{
		return FALSE;
	}

	const String options=a_spSettings.isValid()?
			a_spSettings->catalog<String>("options"): "";
	const BWORD binary=options.match(".*\\bbinary\\b.*");

	RecordView::saveRecordGroup(m_spScope,m_spRecordGroup,a_filename,binary);
	return TRUE;
}

void SurfaceAccessibleRecord::groupNames(Array<String>& a_rNameArray,
	SurfaceAccessibleI::Element a_element) const
{
	a_rNameArray.clear();

	//* TODO
}

void SurfaceAccessibleRecord::attributeSpecs(
	Array<SurfaceAccessibleI::Spec>& a_rSpecs,
	SurfaceAccessibleI::Element a_element) const
{
//	feLog("SurfaceAccessibleRecord::attributeSpecs\n");

	a_rSpecs.clear();

	const String layoutName=elementLayout(a_element);
	sp<Layout> spElementLayout=m_spScope->lookupLayout(layoutName);
	sp<TypeMaster> spTypeMaster=registry()->master()->typeMaster();

//	feLog("SurfaceAccessibleRecord::attributeSpecs layout %s %p\n",
//			layoutName.c_str(),spElementLayout.raw());

	if(spElementLayout.isValid())
	{
		const U32 attrCount=spElementLayout->attributeCount();

//		feLog("SurfaceAccessibleRecord::attributeSpecs attrCount %d\n",
//				attrCount);

		for(U32 attrIndex=0;attrIndex<attrCount;attrIndex++)
		{
			sp<fe::Attribute> spAttribute=spElementLayout->attribute(attrIndex);
			const String attrName=spAttribute->name();

			if(attrName==":CT" || attrName==":SN")
			{
				continue;
			}

			const String attrTypeName=spTypeMaster->lookupBaseType(
					spAttribute->type());
			String typeName;

			if(attrTypeName=="string")
			{
				typeName="string";
			}
			else if(attrTypeName=="I32")
			{
				typeName="integer";
			}
			else if(attrTypeName=="F32")
			{
				typeName="real";
			}
			else if(attrTypeName=="vector3f")
			{
				typeName="vector3";
			}

//			feLog("SurfaceAccessibleRecord::attributeSpecs %d/%d %s %s %s\n",
//					attrIndex,attrCount,
//					attrTypeName.c_str(),typeName.c_str(),attrName.c_str());

			SurfaceAccessibleI::Spec spec;
			spec.set(commonName(attrName),typeName);

			a_rSpecs.push_back(spec);
		}
	}
}

void SurfaceAccessibleRecord::require(SurfaceAccessibleI::Element a_element,
	String a_attrName,String a_attrTypeName)
{
	const String layoutName=elementLayout(a_element);

	m_spScope->declare(layoutName);
	m_spScope->support(a_attrName,a_attrTypeName);
	m_spScope->populate(layoutName,a_attrName);
}

sp<SurfaceAccessorI> SurfaceAccessibleRecord::accessor(
	Element a_element,String a_name,Creation a_create)
{
	FEASSERT(m_weakRecord.isValid());

	sp<SurfaceAccessorRecord> spAccessor(new SurfaceAccessorRecord);

	if(spAccessor.isValid())
	{
		FEASSERT(m_weakRecord.isValid());

		spAccessor->bind(m_weakRecord);
		spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));

		if(spAccessor->bind(a_element,a_name) ||
				a_create==SurfaceAccessibleI::e_createMissing)
		{
			return spAccessor;
		}
	}

	return sp<SurfaceAccessorI>(NULL);
}

sp<SurfaceAccessorI> SurfaceAccessibleRecord::accessor(
	Element a_element,Attribute a_attribute,Creation a_create)
{
	if(!m_weakRecord.isValid())
	{
		feX(e_notInitialized,"no data");
	}

	FEASSERT(m_weakRecord.isValid());

	sp<SurfaceAccessorRecord> spAccessor(new SurfaceAccessorRecord);

	if(spAccessor.isValid())
	{
		FEASSERT(m_weakRecord.isValid());

		spAccessor->bind(m_weakRecord);

		FEASSERT(m_weakRecord.isValid());

		spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));

		FEASSERT(m_weakRecord.isValid());

		if(spAccessor->bind(a_element,a_attribute) ||
				a_create==SurfaceAccessibleI::e_createMissing)
		{
			FEASSERT(m_weakRecord.isValid());

			return spAccessor;
		}

//		feLog("SurfaceAccessibleRecord::accessor"
//				" failed element %d attr %d create %d\n",
//				a_element,a_attribute,a_create);
	}

	FEASSERT(m_weakRecord.isValid());

	return sp<SurfaceAccessorI>(NULL);
}

I32 SurfaceAccessibleRecord::count(SurfaceAccessibleI::Element a_element) const
{
//	feLog("SurfaceAccessibleRecord::count element %d\n",a_element);

	switch(a_element)
	{
		case e_point:
		{
			sp<SurfaceAccessorI> spAccessor=
					const_cast<SurfaceAccessibleRecord*>(this)->accessor(
					e_point,e_position);

			if(spAccessor.isNull())
			{
				feLog("SurfaceAccessibleRecord::count point accessor null\n");
				return 0;
			}

			return spAccessor->count();
		}
		case e_primitive:
		{
			sp<SurfaceAccessorI> spAccessor=
					const_cast<SurfaceAccessibleRecord*>(this)->accessor(
					e_primitive,e_vertices);

			if(spAccessor.isNull())
			{
				feLog("SurfaceAccessibleRecord::count primitive accessor null\n");
				return 0;
			}

			return spAccessor->count();
		}
		default:
			;
	}

	return 0;
}

BWORD SurfaceAccessibleRecord::discard(SurfaceAccessibleI::Element a_element,
	SurfaceAccessibleI::Attribute a_attribute)
{
	String name;
	switch(a_attribute)
	{
		case SurfaceAccessibleI::e_position:
			name="P";
			break;
		case SurfaceAccessibleI::e_normal:
			name="N";
			break;
		case SurfaceAccessibleI::e_uv:
			name="uv";
			break;
		default:
			return FALSE;
	}

	return discard(a_element,name);
}

sp<SurfaceI> SurfaceAccessibleRecord::surface(String a_group)
{
//	feLog("SurfaceAccessibleRecord::surface(\"%s\")\n",a_group.c_str());

	//* TODO select form
	//* TODO group
	sp<SurfaceI> spSurface;

	FEASSERT(registry().isValid());

	if(registry().isValid())
	{
		sp<SurfaceAccessorI> spInputProperties=
				accessor(e_primitive,e_properties);

#if FALSE
		const BWORD openCurve=TRUE;		//* HACK
#else
		const BWORD openCurve=(spInputProperties->count() &&
				spInputProperties->integer(0,
				SurfaceAccessibleI::e_openCurve));
#endif

		if(openCurve)
		{
			spSurface=registry()->create("*.SurfaceCurvesAccessible");

			sp<SurfaceCurvesAccessible> spSurfaceCurvesAccessible=
					spSurface;
			if(spSurfaceCurvesAccessible.isValid())
			{
				spSurfaceCurvesAccessible->bind(
						sp<SurfaceAccessibleI>(this));
			}
		}
		else
		{
			spSurface=registry()->create("*.SurfaceTrianglesAccessible");

			sp<SurfaceTrianglesAccessible> spSurfaceTrianglesAccessible=
					spSurface;
			if(spSurfaceTrianglesAccessible.isValid())
			{
				spSurfaceTrianglesAccessible->bind(
						sp<SurfaceAccessibleI>(this));
			}
		}
	}
	else
	{
		feLog("SurfaceAccessibleRecord::surface registry is NULL\n");
	}

	if(spSurface.isNull())
	{
		feLog("SurfaceAccessibleRecord::surface failed to access surface\n");
	}

	return spSurface;
}

sp<SurfaceI> SurfaceAccessibleRecord::subSurface(U32 a_subIndex,String a_group)
{
	//* TODO

	return sp<SurfaceI>(NULL);
}

sp<SpannedRange> SurfaceAccessibleRecord::atomize(
	AtomicChange a_atomicChange,String a_group,U32 a_desiredCount)
{
#if FE_SAR_DEBUG
	feLog("SurfaceAccessibleRecord::atomize\n");
#endif

	if(a_atomicChange==e_pointsOfPrimitives)
	{
		const BWORD checkPages=TRUE;
		return atomizeConnectivity(a_atomicChange,a_group,
				a_desiredCount,checkPages);
	}

	//* fallback to non-atomic

	const BWORD asPoints=(a_atomicChange==e_pointsOnly);
	const Element element=asPoints? e_point: e_primitive;
	sp<SurfaceAccessorI> spAccessor=accessor(element,e_generic);
	const I32 elementCount=spAccessor.isValid()? spAccessor->count(): 0;
	sp<SpannedRange> spSpannedRange(new SpannedRange());

	//* use everything

	if(a_group.empty())
	{
		if(elementCount>0)
		{
			spSpannedRange->nonAtomic().add(0,elementCount-1);
		}
		return spSpannedRange;
	}

	//* use group

//	feLog("SurfaceAccessibleRecord::atomize group \"%s\"\n",a_group.c_str());

	SpannedRange::MultiSpan& rMultiSpan=spSpannedRange->nonAtomic();

#if FALSE
	I32 spanStart= -1;
	for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
	{
		const I32 elementPage=
				asPoints? pointPage(elementIndex): primitivePage(elementIndex);
		if(spanStart<0 && elementPage>=0)
		{
			//* begin span
			spanStart=elementIndex;
		}
		else if(spanStart>=0 && elementPage<0)
		{
			//* complete span
			rMultiSpan.add(spanStart,elementIndex-1);
			spanStart=-1;
		}
	}

	//* complete last span
	if(spanStart>=0)
	{
		rMultiSpan.add(spanStart,elementCount-1);

	}
#else
	const Element groupElement=asPoints? e_pointGroup: e_primitiveGroup;

	sp<SurfaceAccessorI> spElementGroup=accessor(groupElement,a_group);

	spElementGroup->appendGroupSpans(rMultiSpan);
#endif

	return spSpannedRange;
}

sp<SpannedRange> SurfaceAccessibleRecord::atomizeConnectivity(
	AtomicChange a_atomicChange,String a_group,U32 a_desiredCount,
	BWORD a_checkPages)
{
	if(a_checkPages)
	{
		preparePaging(a_atomicChange,a_group);
	}

	sp<SpannedRange> spSpannedRange(new SpannedRange());

	sp<SurfaceAccessorI> spPoints=accessor(e_point,e_position);
	sp<SurfaceAccessorI> spVertices=accessor(e_primitive,e_vertices);

	const U32 pointCount=spPoints->count();
	const U32 primitiveCount=spVertices->count();

	Array<U32> pointPatch(pointCount,0);
	Array<I32> primitivePatch(primitiveCount,0);

	Array<U32> pagePatch(primitiveCount+1,0);	//* pages <= prims+1
	Array<U32> histogram(primitiveCount+1,0);	//* patches <= prims+1

	U32 maxPatch=0;
	U32 maxPage=0;

	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		if(a_checkPages && primitivePage(primitiveIndex)<0)
		{
			FEASSERT(primitiveIndex<primitiveCount);
			primitivePatch[primitiveIndex]= -1;
			continue;
		}

		U32 lowPatch=0;
		U32 lowPage=0;
		BWORD multiPatch=FALSE;
		BWORD zeroPatches=FALSE;
		BWORD multiPage=FALSE;

		const U32 subCount=spVertices->subCount(primitiveIndex);
		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const U32 pointIndex=spVertices->integer(primitiveIndex,subIndex);
			FEASSERT(pointIndex<pointCount);
			U32& rPatch=pointPatch[pointIndex];
			FEASSERT(rPatch<=maxPatch);

			const I32 page=a_checkPages? pointPage(pointIndex): 0;
			if(page>0)
			{
				FEASSERT(page<=I32(primitiveCount));
				const U32 existingPatch=pagePatch[page];
				FEASSERT(existingPatch<=maxPatch);

				if(existingPatch)
				{
					rPatch=existingPatch;
				}
				if(!lowPage)
				{
					lowPage=page;
				}
				else if(I32(lowPage)!=page)
				{
					multiPage=TRUE;
					if(I32(lowPage)>page)
					{
						lowPage=page;
					}
				}
			}
			if(rPatch)
			{
				if(!lowPatch)
				{
					lowPatch=rPatch;
				}
				else if(lowPatch!=rPatch)
				{
					multiPatch=TRUE;
					if(lowPatch>rPatch)
					{
						lowPatch=rPatch;
					}
				}
			}
			else
			{
				zeroPatches=TRUE;
			}
		}

		if(multiPage)
		{
			//* leave for postAtomic
			lowPatch=0;
		}
		else if(multiPatch)
		{
//			feLog("multiPatch\n");

			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const U32 pointIndex=
						spVertices->integer(primitiveIndex,subIndex);
				FEASSERT(pointIndex<pointCount);
				const U32 patch=pointPatch[pointIndex];

//				feLog("prim %d sub %d/%d patch %d\n",
//						primitiveIndex,subIndex,subCount,patch);

				if(patch<1)
				{
//					feLog("assign patch %d to lowPatch %d\n",patch,lowPatch);

					pointPatch[pointIndex]=lowPatch;
				}
				else if(patch!=lowPatch)
				{
//					feLog("connected patch %d != lowPatch %d\n",patch,lowPatch);

					for(U32 pointOther=0;pointOther<pointCount;pointOther++)
					{
						FEASSERT(pointOther<pointCount);
						U32& rOtherPatch=pointPatch[pointOther];

						if(rOtherPatch==patch)
						{
//							feLog("%d merge patch %d into %d\n",
//									pointIndex,rOtherPatch,lowPatch);

							//* merge all 'patch' into 'lowPatch'
							rOtherPatch=lowPatch;
						}
						else if(rOtherPatch==maxPatch)
						{
//							feLog("%d shift patch %d into %d\n",
//									pointIndex,rOtherPatch,patch);

							//* shift all 'maxPatch' into emptied 'patch'
							rOtherPatch=patch;
						}
					}

					for(U32 primitiveOther=0;primitiveOther<primitiveIndex;
							primitiveOther++)
					{
						I32& rOtherPrim=primitivePatch[primitiveOther];
						primitivePatch[primitiveIndex]=lowPatch;
						if(rOtherPrim==I32(patch))
						{
//							feLog("%d merge patch %d into %d\n",
//									primitiveOther,rOtherPrim,lowPatch);

							//* merge all 'patch' into 'lowPatch'
							rOtherPrim=lowPatch;
						}
						else if(rOtherPrim==I32(maxPatch))
						{
//							feLog("%d shift prim %d into %d\n",
//									primitiveOther,rOtherPrim,patch);

							//* shift all 'maxPatch' into emptied 'patch'
							rOtherPrim=patch;
						}
					}

					histogram[lowPatch]+=histogram[patch];
					histogram[patch]=histogram[maxPatch];

					FEASSERT(maxPatch);
					maxPatch--;

//					feLog("maxPatch -> %d\n",maxPatch);
				}
			}
		}
		else if(!lowPatch)
		{
			maxPatch++;
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const U32 pointIndex=
						spVertices->integer(primitiveIndex,subIndex);
				FEASSERT(pointIndex<pointCount);
				pointPatch[pointIndex]=maxPatch;
			}
			lowPatch=maxPatch;
		}
		else if(zeroPatches)
		{
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const U32 pointIndex=
						spVertices->integer(primitiveIndex,subIndex);
				FEASSERT(pointIndex<pointCount);
				pointPatch[pointIndex]=lowPatch;
			}
		}

//		feLog("primitive %d/%d lowPatch %d/%d\n",
//				primitiveIndex,primitiveCount,lowPatch,maxPatch);

		FEASSERT(lowPatch<=maxPatch);

		FEASSERT(primitiveIndex<primitiveCount);
		primitivePatch[primitiveIndex]=lowPatch;

		FEASSERT(lowPatch<=primitiveCount);
		histogram[lowPatch]+=subCount;

		if(lowPatch)
		{
			FEASSERT(lowPage<=primitiveCount);
			pagePatch[lowPage]=lowPatch;
		}
		if(maxPage<lowPage)
		{
			maxPage=lowPage;
		}
	}

#if FE_SAR_DEBUG
	feLog("SurfaceAccessibleRecord::atomizeConnectivity maxPage %d\n",maxPage);
	for(U32 pageIndex=0;pageIndex<=maxPage;pageIndex++)
	{
		FEASSERT(pageIndex<=primitiveCount);
		feLog("page %d uses patch %d\n",pageIndex,pagePatch[pageIndex]);
	}
	feLog("SurfaceAccessibleRecord::atomizeConnectivity maxPatch %d\n",maxPatch);
	for(U32 patchIndex=0;patchIndex<=maxPatch;patchIndex++)
	{
		FEASSERT(patchIndex<=primitiveCount);
		feLog("patch %d has %d points\n",patchIndex,histogram[patchIndex]);
	}
#endif

	const U32 desiredSize=a_desiredCount? primitiveCount/a_desiredCount: 1;

	Array<U32> patchAtom(maxPatch+1,0);

	I32 maxAtom=1;
	U32 atomSize=0;
	for(U32 patchIndex=1;patchIndex<=maxPatch;patchIndex++)
	{
		if(atomSize>=desiredSize)
		{
			maxAtom++;
			atomSize=0;
		}
		FEASSERT(patchIndex<=maxPatch);
		patchAtom[patchIndex]=maxAtom;

		FEASSERT(patchIndex<=primitiveCount);
		atomSize+=histogram[patchIndex];
#if FE_SAR_DEBUG
		feLog("patch %d atom %d\n",patchIndex,maxAtom);
#endif
	}

	Array<SpannedRange::MultiSpan*> multiSpans(maxAtom+1,NULL);
	I32 spanAtom= -1;
	I32 spanStart= -1;
#if FE_SAR_DEBUG
	feLog("first %d from %d\n",spanAtom,spanStart);
#endif
	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		FEASSERT(primitiveIndex<primitiveCount);
		const I32 patchIndex=primitivePatch[primitiveIndex];

		FEASSERT(patchIndex<0 || patchIndex<=I32(maxPatch));
		const I32 atomIndex=patchIndex>=0? patchAtom[patchIndex]: -1;

		FEASSERT(atomIndex<=maxAtom);
		FEASSERT(spanAtom<=maxAtom);
		if(atomIndex!=spanAtom)
		{
			//* complete existing span
			SpannedRange::MultiSpan* pMultiSpan=NULL;
			if(spanAtom>=0)
			{
				FEASSERT(spanAtom<=maxAtom);
				pMultiSpan=multiSpans[spanAtom];
				if(!pMultiSpan)
				{
					pMultiSpan=spanAtom? &spSpannedRange->addAtomic():
							&spSpannedRange->postAtomic();
					multiSpans[spanAtom]=pMultiSpan;
				}
			}
			if(spanStart>=0)
			{
#if FE_SAR_DEBUG
				feLog("complete %d span %d %d\n",
						spanAtom,spanStart,primitiveIndex-1);
#endif
				pMultiSpan->add(spanStart,primitiveIndex-1);
				spanStart= -1;
			}

			//* start another span
			spanAtom=atomIndex;
			if(atomIndex>=0)
			{
				spanStart=primitiveIndex;

#if FE_SAR_DEBUG
				feLog("start %d from %d\n",spanAtom,spanStart);
#endif
			}
		}
	}

	//* complete last span
	SpannedRange::MultiSpan* pMultiSpan=NULL;
	if(spanStart>=0)
	{
		FEASSERT(spanAtom<=maxAtom);
		pMultiSpan=multiSpans[spanAtom];
		if(!pMultiSpan)
		{
			pMultiSpan=spanAtom? &spSpannedRange->addAtomic():
					&spSpannedRange->postAtomic();
			multiSpans[spanAtom]=pMultiSpan;
		}
		pMultiSpan->add(spanStart,primitiveCount-1);
	}

#if FE_SAR_DEBUG
	feLog("brief: %s\n",spSpannedRange->brief().c_str());
	feLog("dump:\n%s\n",spSpannedRange->dump().c_str());
#endif

	return spSpannedRange;
}

} // namespace
