/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "surface.pmh"

namespace fe
{

sp<RecordArray> SurfaceAccessorRecord::recordArray(void) const
{
	if(m_spRecordArray.isValid())
	{
		return m_spRecordArray;
	}

	RecordGroup::iterator it=m_spRG->begin();
	if(it!=m_spRG->end())
	{
		sp<RecordArray> spRecordArray= *it;

		if(!m_surfaceElementRV.record().isValid())
		{
			const_cast<SurfacePoint*>(&m_surfaceElementRV)->bind(spRecordArray);
		}

		return spRecordArray;
	}
	return sp<RecordArray>(NULL);
}

U32 SurfaceAccessorRecord::count(void) const
{
	sp<RecordArray> spRecordArray=recordArray();

	return spRecordArray.isValid()? spRecordArray->length(): 0;
}

U32 SurfaceAccessorRecord::subCount(U32 a_index) const
{
	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
		sp<RecordArray> spRecordArray=recordArray();
		if(spRecordArray.isValid())
		{
			Record record=spRecordArray->getRecord(a_index);
			const Array<I32>& rVertexArray=m_vertexIndexAccessor(record);
			return rVertexArray.size();
		}
	}

	return 0;
}

void SurfaceAccessorRecord::ensureType(String a_attrType)
{
//	feLog("SurfaceAccessorRecord::ensureType \"%s\" vs \"%s\" for \"%s\"\n",
//			a_attrType.c_str(),m_attrType.c_str(),m_attrName.c_str());

	if(a_attrType==m_attrType)
	{
		return;
	}
	if(m_attrType.empty())
	{
		sp<RecordArray> spRecordArray=recordArray();
		if(spRecordArray.isNull())
		{
			feLog("SurfaceAccessorRecord::ensureType(\"%s\")"
					" no records for \"%s\"\n",
					a_attrType.c_str(),m_attrName.c_str());
			return;
		}

		m_attrType=a_attrType;

		sp<Layout> spLayout=spRecordArray->layout();
		const String layoutName=spLayout->name();

//		feLog("SurfaceAccessorRecord::ensureType \"%s\" for \"%s\" in \"%s\"\n",
//				a_attrType.c_str(),m_attrName.c_str(),layoutName.c_str());

		sp<Scope> spScope=spLayout->scope();

//		m_spSurfaceAccessibleI->lock();

		spScope->declare(layoutName);
		spScope->support(m_attrName,m_attrType);
		spScope->populate(layoutName,m_attrName);

		if(m_attrType=="string")
		{
			m_accessorString.setup(spScope,m_attrName,m_attrType);
		}
		else if(m_attrType=="integer")
		{
			m_accessorInteger.setup(spScope,m_attrName,m_attrType);
		}
		else if(m_attrType=="real")
		{
			m_accessorReal.setup(spScope,m_attrName,m_attrType);
		}
		else if(m_attrType=="vector3")
		{
			m_accessorSpatialVector.setup(spScope,m_attrName,m_attrType);
		}

//		m_spSurfaceAccessibleI->unlock();

		return;
	}

#if FE_CODEGEN<=FE_DEBUG
	feLog("SurfaceAccessorRecord::ensureType(\"%s\")"
			" already \"%s\" for \"%s\"\n",
			a_attrType.c_str(),m_attrType.c_str(),m_attrName.c_str());
#endif
}

void SurfaceAccessorRecord::set(U32 a_index,U32 a_subIndex,String a_string)
{
	ensureType("string");

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull() || a_index>=spRecordArray->length())
	{
		return;
	}

	Record record=spRecordArray->getRecord(a_index);
	if(record.isValid())
	{
		m_accessorString(record)=a_string;
	}
}

String SurfaceAccessorRecord::string(U32 a_index,U32 a_subIndex)
{
	ensureType("string");

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull() || a_index>=spRecordArray->length())
	{
		return "";
	}

	Record record=spRecordArray->getRecord(a_index);
	if(record.isValid())
	{
		return m_accessorString(record);
	}

	return "";
}

void SurfaceAccessorRecord::set(U32 a_index,U32 a_subIndex,I32 a_integer)
{
	if(m_attribute==SurfaceAccessibleI::e_properties)
	{
		//* HACK only e_openCurve
		if(a_subIndex!=SurfaceAccessibleI::e_openCurve)
		{
			return;
		}
	}

	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
		FEASSERT(m_element==SurfaceAccessibleI::e_primitive);

		ensureType("intarray");

		sp<RecordArray> spRecordArray=recordArray();
		if(spRecordArray.isNull() || a_index>=spRecordArray->length())
		{
			return;
		}

		Record record=spRecordArray->getRecord(a_index);
		Array<I32>& rVertexArray=m_vertexIndexAccessor(record);

		if(a_subIndex<rVertexArray.size())
		{
			rVertexArray[a_subIndex]=a_integer;
		}

		return;
	}

	ensureType("integer");

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull() || a_index>=spRecordArray->length())
	{
		return;
	}

	Record record=spRecordArray->getRecord(a_index);
	if(record.isValid())
	{
		m_accessorInteger(record)=a_integer;
	}
}

I32 SurfaceAccessorRecord::integer(U32 a_index,U32 a_subIndex)
{
	if(m_attribute==SurfaceAccessibleI::e_properties)
	{
		//* HACK only e_openCurve
		if(a_subIndex!=SurfaceAccessibleI::e_openCurve)
		{
			return 0;
		}
	}

	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
		FEASSERT(m_element==SurfaceAccessibleI::e_primitive);

		ensureType("intarray");

		sp<RecordArray> spRecordArray=recordArray();
		if(spRecordArray.isNull() || a_index>=spRecordArray->length())
		{
			return 0;
		}

		Record record=spRecordArray->getRecord(a_index);
		const Array<I32>& rVertexArray=m_vertexIndexAccessor(record);

		if(a_subIndex>=rVertexArray.size())
		{
			return 0;
		}

		return rVertexArray[a_subIndex];
	}

	ensureType("integer");

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull() || a_index>=spRecordArray->length())
	{
		return 0;
	}

	Record record=spRecordArray->getRecord(a_index);
	if(record.isValid())
	{
		return m_accessorInteger(record);
	}

	return 0;
}

void SurfaceAccessorRecord::append(U32 a_index,I32 a_integer)
{
	ensureType("intarray");

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull())
	{
		feLog("SurfaceAccessorRecord::append(U32,I32) null record array\n");
		return;
	}

	Record record=spRecordArray->getRecord(a_index);
	m_vertexIndexAccessor(record).push_back(a_integer);
}

I32 SurfaceAccessorRecord::append(SurfaceAccessibleI::Form a_form)
{
	const String layoutName=SurfaceAccessibleBase::elementLayout(m_element);

	sp<Scope> spScope;

	if(m_spLayout.isValid())
	{
		spScope=m_spLayout->scope();
	}
	else
	{
		spScope=m_weakRecord.layout()->scope();
		spScope->declare(layoutName);
		m_spLayout=spScope->lookupLayout(layoutName);
	}

	if(m_spLayout.isNull())
	{
		feLog("SurfaceAccessorRecord::append"
				" no such layout \"%s\" for \"%s\" type \"%s\"\n",
				layoutName.c_str(),m_attrName.c_str(),m_attrType.c_str());
		FEASSERT(FALSE);

		return 0;
	}

	Record record=spScope->createRecord(m_spLayout);
	m_spRG->add(record);

	if(m_spRecordArray.isValid())
	{
		return m_spRecordArray->length()-1;
	}

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isValid())
	{
		if(!m_surfaceElementRV.record().isValid())
		{
			m_surfaceElementRV.bind(spRecordArray);
		}

		m_spRecordArray=spRecordArray;

		return spRecordArray->length()-1;
	}

	return 0;
}

void SurfaceAccessorRecord::set(U32 a_index,U32 a_subIndex,Real a_real)
{
	ensureType("real");

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull() || a_index>=spRecordArray->length())
	{
		return;
	}

	Record record=spRecordArray->getRecord(a_index);
	if(record.isValid())
	{
		m_accessorReal(record)=a_real;
	}
}

Real SurfaceAccessorRecord::real(U32 a_index,U32 a_subIndex)
{
	ensureType("real");

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull() || a_index>=spRecordArray->length())
	{
		return Real(0);
	}

	Record record=spRecordArray->getRecord(a_index);
	if(record.isValid())
	{
		return m_accessorReal(record);
	}

	return Real(0);
}

void SurfaceAccessorRecord::set(U32 a_index,U32 a_subIndex,
	const SpatialVector& a_vector)
{
	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
		const I32 pointIndex=integer(a_index,a_subIndex);
		m_surfacePointRV.setIndex(pointIndex);

		m_surfacePointRV.at()=a_vector;
		return;
	}

	ensureType("vector3");

	m_surfaceElementRV.setIndex(a_index);

	switch(m_attribute)
	{
		case SurfaceAccessibleI::e_generic:
			break;
		case SurfaceAccessibleI::e_position:
			m_surfaceElementRV.at()=a_vector;
			return;
		case SurfaceAccessibleI::e_normal:
			m_surfaceElementRV.up()=a_vector;
			return;
		case SurfaceAccessibleI::e_uv:
			m_surfaceElementRV.uvw()=a_vector;
			return;
		case SurfaceAccessibleI::e_color:
			m_surfaceElementRV.color()=a_vector;
			return;
		default:
			return;
	}

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull() || a_index>=spRecordArray->length())
	{
		return;
	}

	Record record=spRecordArray->getRecord(a_index);
	if(record.isValid())
	{
		m_accessorSpatialVector(record)=a_vector;
	}
}

SpatialVector SurfaceAccessorRecord::spatialVector(U32 a_index,U32 a_subIndex)
{
	FEASSERT(m_spSurfaceAccessibleI.isValid());

	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
		const I32 pointIndex=integer(a_index,a_subIndex);
		m_surfacePointRV.setIndex(pointIndex);

		return m_surfacePointRV.at();
	}

#if FALSE
	ensureType("vector3");
#else
	if(m_attrType.empty())
	{
		m_attrType="vector3";

		sp<RecordArray> spRecordArray=recordArray();
		sp<Layout> spLayout=spRecordArray->layout();
		sp<Scope> spScope=spLayout->scope();

		m_accessorSpatialVector.setup(spScope,m_attrName,m_attrType);
	}
#endif

	m_surfaceElementRV.setIndex(a_index);

	switch(m_attribute)
	{
		case SurfaceAccessibleI::e_generic:
			break;
		case SurfaceAccessibleI::e_position:
			return m_surfaceElementRV.at();
		case SurfaceAccessibleI::e_normal:
			return m_surfaceElementRV.up();
		case SurfaceAccessibleI::e_uv:
			return m_surfaceElementRV.uvw();
		case SurfaceAccessibleI::e_color:
			return m_surfaceElementRV.color();
		default:
			return SpatialVector(Real(0),Real(0),Real(0));
	}

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull() || a_index>=spRecordArray->length())
	{
		return SpatialVector(Real(0),Real(0),Real(0));
	}

	Record record=spRecordArray->getRecord(a_index);
	if(record.isValid())
	{
		return m_accessorSpatialVector(record);
	}

	return SpatialVector(Real(0),Real(0),Real(0));
}

BWORD SurfaceAccessorRecord::spatialVector(SpatialVector* a_pVectorArray,
		const Vector2i* a_pIndexArray,I32 a_arrayCount)
{
	for(I32 arrayIndex=0;arrayIndex<a_arrayCount;arrayIndex++)
	{
		const Vector2i& rIndexPair=a_pIndexArray[arrayIndex];

		a_pVectorArray[arrayIndex]=spatialVector(rIndexPair[0],rIndexPair[1]);
	}

	return TRUE;
}

BWORD SurfaceAccessorRecord::bindInternal(SurfaceAccessibleI::Element a_element,
		const String& a_name)
{
	m_element=a_element;
	m_attrName=a_name;

	//* TODO handle actual groups

	if(m_element==SurfaceAccessibleI::e_pointGroup)
	{
		if(a_name.empty())
		{
			m_element=SurfaceAccessibleI::e_point;
			m_attribute=SurfaceAccessibleI::e_position;
			m_attrName="spc:at";
		}
		else
		{
			feLog("SurfaceAccessorRecord::bindInternal"
					" point groups not implemented (\"%s\")\n",
					a_name.c_str());
		}
	}

	if(m_element==SurfaceAccessibleI::e_primitiveGroup)
	{
		if(a_name.empty())
		{
			m_element=SurfaceAccessibleI::e_primitive;
			m_attribute=SurfaceAccessibleI::e_vertices;
			m_attrName="surf:verts";
		}
		else
		{
			feLog("SurfaceAccessorRecord::bindInternal"
					" primitive groups not implemented (\"%s\")\n",
					a_name.c_str());
		}
	}


	FEASSERT(m_weakRecord.isValid());

	if(!m_weakRecord.isValid())
	{
		feLog("SurfaceAccessorRecord::bindInternal invalid Record\n");
		return FALSE;
	}

	//* RecordView bind to Record is not weak
	//* Record ref counting is not thread-safe, apparently
//	m_spSurfaceAccessibleI->lock();

	SurfaceModel surfaceModelRV;
	surfaceModelRV.bind(m_weakRecord);

	FEASSERT(m_weakRecord.isValid());

	if(m_attrName=="surf:verts")
	{
		m_spRG=surfaceModelRV.surfacePointRG();

		sp<RecordArray> spRecordArrayPoint=recordArray();
		if(spRecordArrayPoint.isValid())
		{
			m_surfacePointRV.bind(spRecordArrayPoint);
		}
	}

	switch(m_element)
	{
		case SurfaceAccessibleI::e_point:
			m_spRG=surfaceModelRV.surfacePointRG();
			break;
		case SurfaceAccessibleI::e_vertex:
			m_spRG=surfaceModelRV.surfaceVertexRG();
			break;
		case SurfaceAccessibleI::e_primitive:
			m_spRG=surfaceModelRV.surfacePrimitiveRG();
			break;
		case SurfaceAccessibleI::e_detail:
			m_spRG=surfaceModelRV.surfaceDetailRG();
			break;
		default:
			feLog("SurfaceAccessorRecord::bindInternal invalid element\n");
			return FALSE;
	}

	surfaceModelRV.unbind();

//	m_spSurfaceAccessibleI->unlock();

	FEASSERT(m_weakRecord.isValid());

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isValid())
	{
		m_surfaceElementRV.bind(spRecordArray);

		//* TODO check other attributes
		if(m_attribute==SurfaceAccessibleI::e_position &&
				!m_surfaceElementRV.at.check())
		{
//			feLog("SurfaceAccessorRecord::bindInternal no points\n");
			return FALSE;
		}
		if(m_attribute==SurfaceAccessibleI::e_normal &&
				!m_surfaceElementRV.up.check())
		{
//			feLog("SurfaceAccessorRecord::bindInternal no normals\n");
			return FALSE;
		}
		if(m_attribute==SurfaceAccessibleI::e_uv &&
				!m_surfaceElementRV.uvw.check())
		{
//			feLog("SurfaceAccessorRecord::bindInternal no uvs\n");
			return FALSE;
		}
		if(m_attribute==SurfaceAccessibleI::e_color &&
				!m_surfaceElementRV.color.check())
		{
//			feLog("SurfaceAccessorRecord::bindInternal no colors\n");
			return FALSE;
		}
	}

	m_spRecordArray=spRecordArray;

	FEASSERT(m_weakRecord.isValid());

	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
		sp<Scope> spScope=m_weakRecord.layout()->scope();
		m_vertexIndexAccessor.setup(spScope,"surf:verts","intarray");
	}

	return TRUE;
}

} // namespace
