/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#define	FE_TND_COOK_DEBUG		FALSE
#define	FE_TND_PRINT_MESSAGES	TRUE

#include "terminal/terminal.pmh"

namespace fe
{

TerminalNode::TerminalNode(void)
{
#if	FE_TND_COOK_DEBUG
	feLog("TerminalNode::TerminalNode\n");
#endif

#if FE_COUNTED_TRACK
	Counted::registerRegion(this,sizeof(TerminalNode),"TerminalNode");
#endif
}

TerminalNode::~TerminalNode(void)
{
#if	FE_TND_COOK_DEBUG
	feLog("~TerminalNode\n");
#endif

	clear();

#if FE_COUNTED_TRACK
	Counted::deregisterRegion(this);
#endif
}

void TerminalNode::clear(void)
{
	m_spScope=NULL;

	clearReferences();
}

void TerminalNode::setupOperator(sp<Scope> a_spScope,
	String a_implementation,String a_name)
{
	m_spScope=a_spScope;

	sp<Registry> spRegistry=m_spScope->registry();

	m_spDrawOutput=spRegistry->create("DrawI.TerminalDraw");
	if(m_spDrawOutput.isValid())
	{
		m_spDrawOutput->setName("DrawOutput");
	}

	setupScope(m_spScope);

	m_spOperatorSurfaceI=spRegistry->create(
			"OperatorSurfaceI."+a_implementation);

	if(m_spOperatorSurfaceI.isValid())
	{
		m_spOperatorSurfaceI->setName(a_name);
		insertHandler(m_spOperatorSurfaceI);

		updateCatalog();
	}
}

void TerminalNode::updateCatalog(void)
{
	m_surfaceInputArray.clear();

	sp<Catalog> spCatalog=m_spOperatorSurfaceI;

	//* scan for inputs
	Array<String> keys;
	spCatalog->catalogKeys(keys);

	String firstInput;

	const U32 number=keys.size();
	for(U32 m=0;m<number;m++)
	{
		const String keyname=keys[m];

		Instance instance;
		bool success=spCatalog->catalogLookup(keyname,instance);
		if(!success)
		{
			continue;
		}

		if(instance.is< sp<Component> >())
		{
#if	FE_TND_COOK_DEBUG
			feLog("TerminalNode::updateCatalog \"%s\"\n",keyname.c_str());
#endif

			const String implementation=
					spCatalog->catalogOrDefault<String>(keyname,
					"implementation","SurfaceAccessibleI");

			if(implementation.dotMatch("SurfaceAccessibleI"))
			{
				const String io=spCatalog->catalogOrDefault<String>(
						keyname,"IO","input");

				sp<Component>& rspComponent=instance.cast< sp<Component> >();
				sp<SurfaceAccessibleI> spSurfaceAccessibleI=rspComponent;

				if(io=="input")
				{
					I32 channel=m_surfaceInputArray.size();
					spCatalog->catalog<I32>(keyname,"channel")=channel;
					m_surfaceInputArray.push_back(spSurfaceAccessibleI);
					String lookup;
					lookup.sPrintf("input:%d",channel);
					spCatalog->catalog<String>(lookup)=keyname;

					if(firstInput.empty())
					{
						firstInput=keyname;
					}
				}
				else
				{
					spCatalog->catalog<String>("output:0")=keyname;
				}
			}
		}
	}
	if(spCatalog->catalog<String>("output:0")=="")
	{
		spCatalog->catalog<String>("output:0")="Output Surface";
		spCatalog->catalog<String>("Output Surface","copy")=firstInput;
		spCatalog->catalog<String>("Output Surface","io")="output";
	}

	const String lookup=spCatalog->catalog<String>("output:0");
	m_spSurfaceOutput=spCatalog->catalog< sp<Component> >(lookup);

	sp<TerminalDraw> spTerminalDraw=m_spDrawOutput;
	if(spTerminalDraw.isValid())
	{
		spTerminalDraw->setSurfaceAccessible(m_spSurfaceOutput);
		spCatalog->catalog< sp<Component> >("")=spTerminalDraw;
	}
}

BWORD TerminalNode::cook(Real a_frame)
{
#if	FE_TND_COOK_DEBUG
	feLog("TerminalNode::cook %.6G\n",a_frame);
#endif
	BWORD success=TRUE;

	const double time=0.0;

	const Real startFrame=0.0;
	const Real endFrame=0.0;

	if(m_spOperatorSurfaceI.isValid())
	{
		updateCatalog();

		FEASSERT(m_spSurfaceOutput.isValid());

		sp<Catalog> spCatalog=m_spOperatorSurfaceI;

		m_aFrame(m_cookSignal)=a_frame;
		m_aStartFrame(m_cookSignal)=startFrame;
		m_aEndFrame(m_cookSignal)=endFrame;
		m_aTime(m_cookSignal)=time;
		m_aSurfaceOutput(m_cookSignal)=m_spSurfaceOutput;

		try
		{
			m_spSignalerI->signal(m_cookSignal);
		}
		catch(const Exception& e)
		{
			spCatalog->catalog<String>("error")+=
					"FE EXCEPTION: "+print(e);
			success=FALSE;
		}
		catch(const std::exception& std_e)
		{
			spCatalog->catalog<String>("error")+=
					"stdlib EXCEPTION: "+String(std_e.what());
			success=FALSE;
		}
#ifdef FE_BOOST_EXCEPTIONS
		catch(const boost::exception& boost_e)
		{
			spCatalog->catalog<String>("error")+=
					"boost EXCEPTION: "+
					String(boost::diagnostic_information(boost_e).c_str());
			success=FALSE;
		}
#endif
		catch(...)
		{
			spCatalog->catalog<String>("error")+=
					"UNRECOGNIZED EXCEPTION";
			success=FALSE;
		}

		String messageString=spCatalog->catalogOrDefault<String>("message","");
		if(!messageString.empty())
		{
			spCatalog->catalog<String>("message")="";
#if	FE_TND_PRINT_MESSAGES
			feLog("TerminalNode::cook MESSAGE '%s'\n",
					messageString.c_str());
#endif
		}

		String warningString=spCatalog->catalogOrDefault<String>("warning","");
		if(!warningString.empty())
		{
			spCatalog->catalog<String>("warning")="";
#if	FE_TND_PRINT_MESSAGES
			feLog("[33mTerminalNode::cook WARNING '%s'[0m\n",
					warningString.c_str());
#endif
		}

		String errorString=spCatalog->catalogOrDefault<String>("error","");
		if(!errorString.empty())
		{
			spCatalog->catalog<String>("error")="";
#if	FE_TND_PRINT_MESSAGES
			feLog("[31mTerminalNode::cook ERROR '%s'\n[0m",
					errorString.c_str());
#endif
			success=FALSE;
		}
	}

	m_lastTime=time;

#if	FE_TND_COOK_DEBUG
	feLog("TerminalNode::cook done\n");
#endif

	return success;
}

} // namespace
